# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import List
from typing import Optional

import click
from vdk.api.plugin.hook_markers import hookspec
from vdk.api.plugin.plugin_registry import IPluginRegistry
from vdk.internal.builtin_plugins.run.execution_results import ExecutionResult
from vdk.internal.builtin_plugins.run.execution_results import StepResult
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.builtin_plugins.run.step import Step
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.context import CoreContext


class CoreHookSpecs:
    """
    These are hook specifications that enable plugins to hook at core events during CLI execution.
    Those are hooks applicable for all commands (run, deploy, etc.)
    """

    @hookspec(historic=True)
    def vdk_start(
        self, plugin_registry: IPluginRegistry, command_line_args: List
    ) -> None:
        """
        Called upon immediate start of VDK just after (only) PluginRegistry is initialized.
        Nothing else (including logging, monitoring) is initialized at this point.

        :param plugin_registry: the plugin registry. Use it to register new plugin hooks
        :param command_line_args: the command arguments, can be modified.
        """
        pass

    @hookspec
    def vdk_command_line(self, root_command: click.Group) -> None:
        """
        Customize commands by adding new commands, customizing existing commands or setting default values to cli parameters

        Examples:
        Add new commands:

        .. code-block:: python
            @click.command
            def custom_command(...)
                do_something ...
            command.add_command(custom_command)

        Add options to existing commands:

        .. code-block:: python
            # note this is click.option which is a function (not the class click.Option).
            # You need to set expose_value=False if the original command method does not allow for extra arguments.
            custom_option = click.option('--custom', '-c', expose_value=False, callback = f, ...)
            # get the subcommand we want to add the option
            sub_cmd = root_command.get(None, sub_command_name)
            # replace the command with same one but now decorated with the new option
            root_command.add_command(custom_option(sub_cmd))

        Set default values to parameters:

        .. code-block:: python
            # the hookimpl decorator should have its trylast parameter set to True,
            # so that all command plugins are loaded before defaults are set

            from vdk.api.plugin.plugin_utils import set_defaults_for_specific_command, set_defaults_for_all_commands

            set_defaults_for_all_commands(root_command, {'team': 'team_name'})
            set_defaults_for_specific_command(root_command, 'command_name', {'param1': '1', 'param2': '2'})

        :param root_command: the root command which contains all sub-commands
        """

    @hookspec
    def vdk_configure(self, config_builder: "ConfigurationBuilder") -> None:
        """
        Configuring the application. Add new configuration keys and values.
        Two type of plugins can implement this hook
        * Plugin that need configuration to be provided for them later in Configuration instance and define configuration keys.
          Usually you would decorate that plugin with tryfirst=True so it's executed before all provider plugins

        * Configuration values provider plugins (e.g yaml file reader, environment vraibles configuration, etc.)
          Leave with no decorator or set trylast decorator. For such plugins it's good idea to specify configuration key
          with pattern '__config_provider__{name_of_provider}' with description that tell users on how configuration variables are updated.
          vdk config-help would print those instructions separately

        TODO: we should consider splitting this into two hooks vdk_configure_definine_vars and vdk_configure_providers

        :param config_builder:
        """
        pass

    @hookspec
    def vdk_initialize(self, context: CoreContext) -> None:
        """
        Use it to initialize the CLI or initialize modules

        :param context: The context of the current CLI execution
        """
        pass

    @hookspec
    def vdk_exception(self, exception: Exception) -> bool:
        """
        Called in case the CLI is about to exit with error.

        :return: True if exception is handled and does not need to be logged by main, nor affect the exit code
        """
        pass

    @hookspec
    def vdk_exit(self, context: CoreContext, exit_code: int) -> None:
        """
        Called last just before CLI exits.

        :param context: The context of the current CLI execution
        :param exit_code: The exit code with which the CLI is about to exit.
        """
        pass


class JobRunHookSpecs:
    """
    These are hook specifications that enable plugins to hook during the run (or execution) of a Data Job.
    Generally done with "run" command of the CLI.

    They are called during as part of execution cycle of Data Job or Data Job template.
    Job template execution inherits JobContext of the job that started but all hooks are still called.
    That's because a job template is executed as a normal data job.
    """

    @hookspec
    def initialize_job(self, context: JobContext) -> None:
        """
        Called when a new job is starting to be executed. And will be use to initialize it.
        Default implementations will collect and setup list of steps to execute,
        initialize JobInput interfaces (e.g. initialize Managed databases connections)

        :param context: The Job Context of the current run.
        """
        pass

    @hookspec(firstresult=True)
    def run_job(self, context: JobContext) -> Optional[ExecutionResult]:
        """
        The method that executes the actual run of the data job.

        By default it executes the provided steps in context.steps one by one. It also run last (trylast is set)

        Will stop at first hook implementation that returns non-None result.
        Hence if you want to override the default implementation return a non-None result.
        If you just want to decorate it, then return None (and it's good idea to specify tryfirst=True also)

        E.g
        @hookimpl(tryfirst=True)
        create_and_run_data_job(context): # decorate with logs
            log.info(f"Start {context.name}"
            # no return

        @hookimpl
        create_and_run_data_job(context): # alternative execution algorithm
            results = run_in_parallel(steps)
            return results.all_success

        :param context: the job context
        """
        pass

    @hookspec(firstresult=True)
    def run_step(self, context: JobContext, step: Step) -> Optional[StepResult]:
        """
        Executes a step.

        The default implementation will run "runner_func" only and returns. It also run last (trylast is set)

        Will stop at first hook implementation that returns non-None result.
        Hence if you want to override the default implementation return a non-None result.
        If you just want to decorate it, then return None (and it's good idea to specify tryfirst=True also)

        :param context: the job context
        :param step: the step that will be run
        """
        pass

    @hookspec
    def finalize_job(self, context: JobContext) -> None:
        """
        Job has finished to execute.
        :param context: the job context. Use it to see the job status and the job steps statuses as well.
        """
        pass



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from pluggy import PluginManager


class VdkPluginManager(PluginManager):
    """
    Abstract away direct access (in terms of imports) to pluggy PluginManager.
    """

    pass



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
import pathlib
from abc import ABC
from dataclasses import dataclass
from typing import Callable

from vdk.api.job_input import IJobInput

log = logging.getLogger(__name__)


# The function accept Step (below class) and IJobInput and
# return true if the step has been executed and false if it is not (valid) executable step.
# On error it's expected to raise an exception.
StepFunction = Callable[["Step", IJobInput], bool]


@dataclass
class Step(ABC):
    """
    A step that will be executed when running a data job.
    """

    # the name of the concrete step (e.g name of the file)
    name: str
    # string representing a step type (sql or python).
    type: str
    # function that will execute the actual step
    runner_func: StepFunction
    # file where the step is defined
    file_path: pathlib.Path
    # the root job directory
    job_dir: pathlib.Path
    # parent Step
    parent: Step | None = None


@dataclass
class StepBuilder:
    def __init__(self):
        self.__steps: list[Step] = []

    def add_step(self, step: Step) -> None:
        self.__steps.append(step)

    def get_steps(self) -> list[Step]:
        return self.__steps



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import imp
import inspect
import logging
import pathlib
import sys
from typing import Callable
from typing import List

from vdk.api.job_input import IJobInput
from vdk.internal.builtin_plugins.run.step import Step
from vdk.internal.core import errors

log = logging.getLogger(__name__)

TYPE_SQL = "sql"
TYPE_PYTHON = "python"


class JobFilesLocator:
    """
    Locate the data job files that would be executed by us.
    """

    def get_script_files(self, directory: pathlib.Path) -> List[pathlib.Path]:
        """Locates the files in a directory, that are supported for execution.

        Files supported for execution are: .sql and .py
        Other files in the directory are ignored.

        :return: List of files from the directory that supported for execution, sorted alphabetically by name.
        :rtype: :class:`.list`

        """
        script_files = [
            x
            for x in directory.iterdir()
            if (x.name.lower().endswith(".sql") or x.name.lower().endswith(".py"))
        ]
        script_files.sort(key=lambda x: x.name)
        log.debug(f"Script files of {directory} are {script_files}")
        return script_files


class StepFuncFactory:
    """
    Implementations of runner_func for running steps
    """

    @staticmethod
    def run_sql_step(step: Step, job_input: IJobInput) -> bool:
        """ """
        with open(step.file_path, encoding="utf8") as sql_file:
            sql = sql_file.read()
        job_input.execute_query(sql)
        return True

    @staticmethod
    def run_python_step(step: Step, job_input: IJobInput) -> bool:
        """Runs a Python step script.

        In order for a Python file to be executed as a Job Step, it must follow the contract of having:

        def run(job_input)

        method defined, which is the starting point of the Step.

        :return: True if there was a run(job_input) method to run, False otherwise
        :rtype: :class:`.bool`
        """

        try:
            sys.path.insert(0, str(step.job_dir))
            filename = step.file_path.name
            namespace = "step_" + str(filename).lower().strip(".py")
            success = False

            try:
                log.debug("Loading %s ..." % filename)
                python_module = imp.load_source(namespace, str(step.file_path))
                log.debug("Loading %s SUCCESS" % filename)
            except BaseException as e:
                log.info("Loading %s FAILURE" % filename)
                errors.log_and_rethrow(
                    to_be_fixed_by=errors.ResolvableBy.USER_ERROR,
                    log=log,
                    what_happened="Failed loading job sources of %s" % filename,
                    why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                    consequences=errors.MSG_CONSEQUENCE_TERMINATING_APP,
                    countermeasures=errors.MSG_COUNTERMEASURE_FIX_PARENT_EXCEPTION
                    + " Most likely importing a dependency or data job step failed, see"
                    + " logs for details and fix the failed step (details in stacktrace).",
                    exception=e,
                    wrap_in_vdk_error=True,
                )

            for _, func in inspect.getmembers(python_module, inspect.isfunction):
                if func.__name__ == "run":
                    try:
                        log.info("Entering %s#run(...) ..." % filename)
                        StepFuncFactory.invoke_run_function(func, job_input)
                        success = True
                        return True
                    finally:
                        if success:
                            log.info("Exiting  %s#run(...) SUCCESS" % filename)
                        else:
                            log.error("Exiting  %s#run(...) FAILURE" % filename)
            log.warn(
                "File %s does not contain a valid run() method. Nothing to execute. Skipping %s,"
                + " and continuing with other files (if present).",
                filename,
                filename,
            )
            return success
        finally:
            sys.path.remove(str(step.job_dir))

    @staticmethod
    def invoke_run_function(func: Callable, job_input: IJobInput):
        # https://docs.python.org/3/library/inspect.html#inspect.getfullargspec
        full_arg_spec = inspect.getfullargspec(func)
        parameter_names = full_arg_spec[0]
        #  if 'job_input' not in argument_names: TODO: what should happen ? What is the behaviour in SC
        possible_arguments = {
            "job_input": job_input,
        }
        # enable plugins to add different types of job input
        actual_arguments = {
            arg_name: arg_value
            for arg_name, arg_value in possible_arguments.items()
            if arg_name in parameter_names
        }
        if actual_arguments:
            func(**actual_arguments)
        else:
            errors.log_and_throw(
                to_be_fixed_by=errors.ResolvableBy.USER_ERROR,
                log=log,
                what_happened=f"I'm trying to call method 'run' and failed.",
                why_it_happened=f"Method is missing at least one job input parameter to be passed",
                consequences="Current Step (python file) will fail, and as a result the whole Data Job will fail. ",
                countermeasures="Make sure that you have specified a job input parameter in the signature of the "
                "run method. "
                f"Possible parameters of run function are: {list(possible_arguments.keys())}.",
            )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
"""
Ingestion Utilities
"""
import datetime
import itertools
import logging
import queue
import threading
import uuid
from typing import Any
from typing import List

from vdk.internal.core import errors

log = logging.getLogger(__name__)


class AtomicCounter:
    """
    Thread-safe incrementing counter
    """

    def __init__(self, start: int = 0):
        """Init a new atomic counter to start value"""
        self.value: int = start
        self._lock = threading.Lock()

    def increment(self, step: int = 1) -> int:
        """Atomically increment the counter by step, return new value"""
        self.get_and_increment(step)
        return self.value

    def get_and_increment(self, step: int = 1) -> int:
        """Atomically increment counter by step, return old value"""
        with self._lock:
            old_value: int = self.value
            self.value += step
            return old_value

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return str(self)


def get_page_generator(data, page_size=10000):
    # TODO add support for Pandas
    try:
        while True:
            page = data.fetchmany(page_size)
            if not page:
                return
            yield page
    except AttributeError:
        it = iter(data)
        while True:
            page = list(itertools.islice(it, page_size))
            if not page:
                return
            yield page


def validate_column_count(data: iter, column_names: iter):
    if data:
        if len(column_names) != len(data[0]):
            errors.log_and_throw(
                to_be_fixed_by=errors.ResolvableBy.USER_ERROR,
                log=log,
                what_happened="Failed to post tabular data for ingestion.",
                why_it_happened="The number of column names are not matching the number of values in at least on of"
                "the rows. You provided columns: '{column_names}' and data row: "
                "'{data_row}'".format(column_names=column_names, data_row=data[0]),
                consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                countermeasures="Check the data and the column names you are providing, their count should match.",
            )


def convert_table(table: iter, column_names: iter) -> List[dict]:
    """
    Converts tabular data into dictionary objects

    :param table: iter
       A representation of a two-dimensional array that allows iteration over rows.
    :param column_names: iter
       Names of the table columns.
    :return: list of dicts containing the converted table objects.
    """
    converted_rows = []
    for row in table:
        cdf_row = dict()
        for index, value in enumerate(row):
            value = _handle_special_types(value)
            cdf_row[column_names[index]] = value
        converted_rows.append(cdf_row)

    return converted_rows


def _handle_special_types(value: Any) -> Any:
    """
    Handle data types that require special care to be correctly processed.

    :param value: Any
        The value to be converted. Only `datetime.date` and `uuid.UUID` objects
        are converted. All other data types are returned as is.
    """
    if isinstance(value, datetime.date):
        return value.isoformat()
    if isinstance(value, uuid.UUID):
        return str(
            value
        )  # there is no default json serializer for UUID type, hence cast to string
    return value


def wait_completion(objects_queue: queue.Queue, payloads_queue: queue.Queue):
    objects_queue.join()
    payloads_queue.join()


def is_iterable(obj: Any) -> bool:
    """
    According to Python doc this is the most reliable way to determine if object is iterable.
    See https://docs.python.org/3/library/collections.abc.html#collections.abc.Iterable
    :param obj: the object to check
    :return: true or false :)
    """
    try:
        _ = iter(obj)
        return True
    except TypeError:
        return False



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Any

from vdk.api.job_input import IProperties
from vdk.api.plugin.plugin_input import IPropertiesFactory
from vdk.api.plugin.plugin_input import IPropertiesRegistry
from vdk.internal.builtin_plugins.config.job_config import JobConfigKeys
from vdk.internal.builtin_plugins.job_properties.cached_properties import (
    CachedPropertiesWrapper,
)
from vdk.internal.builtin_plugins.job_properties.datajobs_service_properties import (
    DataJobsServiceProperties,
)
from vdk.internal.builtin_plugins.job_properties.properties_config import (
    PropertiesConfiguration,
)
from vdk.internal.builtin_plugins.job_properties.propertiesnotavailable import (
    PropertiesNotAvailable,
)
from vdk.internal.core import errors
from vdk.internal.core.config import Configuration
from vdk.internal.core.errors import ResolvableBy
from vdk.internal.util.decorators import LogDecorator

log: logging.Logger = logging.getLogger(__name__)


class PropertiesRouter(IPropertiesRegistry, IProperties):
    """
    Implements registry of IProperties.
    """

    def __init__(self, job_name: str, cfg: Configuration):
        self.__properties_builders = {}
        self.__cached_properties_impl = None
        self.__job_name = job_name
        self.__team_name = cfg.get_value(JobConfigKeys.TEAM.value)
        self.__config = PropertiesConfiguration(cfg)

    @LogDecorator(log)
    def set_properties_factory_method(
        self, properties_type: str, properties_factory: IPropertiesFactory
    ) -> None:
        self.__properties_builders[properties_type] = properties_factory

    def get_property(self, name: str, default_value: Any = None) -> str:
        return self.__get_properties_impl().get_property(name, default_value)

    def get_all_properties(self) -> dict:
        return self.__get_properties_impl().get_all_properties()

    def set_all_properties(self, properties: dict):
        self.__get_properties_impl().set_all_properties(properties)

    def has_properties_impl(self) -> bool:
        """
        If any properties implementation backend available
        :return: bool
        """
        return not isinstance(self.__get_properties_impl(), PropertiesNotAvailable)

    def __get_properties_impl(self):
        """
        Singleton properties backend implementation.
        """
        if not self.__cached_properties_impl:
            if self.__properties_builders:
                self.__cached_properties_impl = (
                    self.__setup_properties_from_factory_method()
                )
            else:
                why = "No properties client implementation has been installed or configured. "
                countermeasures = (
                    "Check if you have installed properly vdk. "
                    "Check if you need to install some plugin for properties. "
                    "If a plugin is installed, check out plugin documentation"
                    " whether the plugin is configured correctly."
                )
                error_handler = self.__properties_not_available_error_handler(
                    why=why, countermeasures=countermeasures
                )
                self.__cached_properties_impl = PropertiesNotAvailable(error_handler)
        return self.__cached_properties_impl

    def __choose_from_default_type(self, properties_type: str):
        if properties_type in self.__properties_builders:
            return self.__properties_builders.get(properties_type)
        else:
            errors.log_and_throw(
                errors.ResolvableBy.CONFIG_ERROR,
                log,
                "",
                f"properties default type was configured to be {properties_type} "
                f"no such properties api implementation has been registered",
                f"",
                f"Check if the job has not been mis-configured - for example misspelling error. "
                f"See config-help for help on configuration. Existing properties types are: {list(self.__properties_builders.keys())} "
                f"Alternatively make sure the correct plugin has been installed "
                f"providing the properties api implementation ",
            )

    def __setup_properties_from_factory_method(self):
        try:
            properties_type = self.__config.get_properties_default_type()
            factory_method = None
            if properties_type:
                factory_method = self.__choose_from_default_type(properties_type)
            elif "default" in self.__properties_builders:
                factory_method = self.__choose_from_default_type("default")
            elif len(self.__properties_builders) == 1:
                factory_method = self.__properties_builders[
                    list(self.__properties_builders.keys())[0]
                ]
            else:
                errors.log_and_throw(
                    errors.ResolvableBy.CONFIG_ERROR,
                    log,
                    "Properties API client cannot be chosen.",  # set by handler
                    f"Too many choices for properties client implementation.",
                    f"Properties API functionality does not work.",  # set by handler
                    f"Configure which properties client implementation "
                    f"to use with properties_default_type config option. "
                    f"See config-help for help on configuration. Existing properties types are: {list(self.__properties_builders.keys())}",
                )

            for p in self.__config.get_properties_write_preprocess_sequence():
                if p not in self.__properties_builders.keys():
                    errors.log_and_throw(
                        to_be_fixed_by=ResolvableBy.CONFIG_ERROR,
                        log=log,
                        what_happened=f"A non-valid properties type {p} configured in "
                        "PROPERTIES_WRITE_PREPROCESS_SEQUENCE.",
                        why_it_happened=f"No IPropertiesServiceClient handler for property type {p} "
                        "was registered in IPropertiesRegistry.",
                        consequences="The write pre-processing failed.",
                        countermeasures=f"Either remove the non-valid properties type {p} from "
                        "PROPERTIES_WRITE_PREPROCESS_SEQUENCE configuration, or"
                        "register a corresponding IPropertiesServiceClient "
                        "implementation via set_properties_factory_method API.",
                    )

            service_properties = DataJobsServiceProperties(
                job_name=self.__job_name,
                team_name=self.__team_name,
                properties_service_client=factory_method(),
                write_preprocessors=[
                    self.__properties_builders.get(p)()
                    for p in self.__config.get_properties_write_preprocess_sequence()
                ],
            )
            return CachedPropertiesWrapper(service_properties)
        except Exception as e:
            log.warning(
                "Could not build properties provider. "
                "This will result in failure if getting/setting properties is used."
            )
            error_handler = self.__properties_not_available_error_handler(str(e))
            return PropertiesNotAvailable(error_handler)

    @staticmethod
    def __properties_not_available_error_handler(why, countermeasures=None):
        if not countermeasures:
            countermeasures = " Check why it happened and try to resolve the issue or open a ticket to SRE team."

        error_handler = lambda methodname: errors.log_and_throw(
            to_be_fixed_by=errors.ResolvableBy.CONFIG_ERROR,
            log=logging.getLogger(__name__),
            what_happened="I'm trying to call method '{}' and failed.".format(
                methodname
            ),
            why_it_happened=why,
            consequences="Current  Step will fail, and as a result the whole Data Job will fail.",
            countermeasures=countermeasures,
        )
        return error_handler



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging

log = logging.getLogger(__name__)


class WriteToFileAction:
    """
    Writes a string message to a specified file upon completion of a data job execution.

    This is intended to provide a termination status of a data job in a Kubernetes environment
    (see https://kubernetes.io/docs/tasks/debug-application-cluster/determine-reason-pod-failure/).
    Keep in mind that there is a limit to the length of the termination message.
    """

    def __init__(self, filename, show_log_messages=True):
        self.filename = filename
        self.show_log_messages = show_log_messages

    def write_to_file(self, message) -> None:
        """
        This method is intended to be used by the termination message plugin.
        Writes a message to the configured filename (truncating it first).
        This means that if two or more messages are writen only the latest one
        will be the actual data job termination messsage.

        :param message: Json formatted string with the termination message
        :return: None
        """
        if not self.filename:
            return

        try:
            with open(self.filename, "w") as file:
                file.write(message)
        except OSError as e:
            if self.show_log_messages:
                log.debug(
                    f'Unable to write termination message to file "{self.filename}". {e}'
                )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import inspect
from typing import Any
from typing import List
from typing import Optional

from vdk.internal.core.config import Configuration


def class_fqname(py_object: Any) -> str:
    """
    :param py_object: Any python object
    :return: the full qualified name of the class (package.package.classname)
    """
    module = inspect.getmodule(py_object)
    if module is None:
        return py_object.__class__.__name__
    return f"{py_object.__class__.__module__}.{py_object.__class__.__name__}"


def parse_config_sequence(
    cfg: Configuration, key: str, sep: Optional[str] = None
) -> List:
    """
    Parse a configuration variable, whose value is a string sequence, by a
    specified delimiter.
    :param cfg: Configuration
        The configuration of Versatile Data Kit.
    :param key: string
        The name of the configuration variable that is to be parsed.
    :param sep: string
        Optional. The delimiter by which the configuration variable string is
        to be split. If not specified or set as None, any whitespace will be
        considered a delimiter, and the result from the split will have no
        empty strings.
    :return:
    """
    sequence = cfg.get_value(key)
    if sequence:
        sequence = [i.strip() for i in sequence.split(sep)]
    return sequence if sequence else []




RELEASE_VERSION='0.3.610538876'
BUILD_DATE='Thu Aug 11 15:23:14 UTC 2022'
BUILD_MACHINE_INFO='Linux runner-yxfxmkoy-project-28359933-concurrent-0cqtfz 5.4.181-99.354.amzn2.x86_64 #1 SMP Wed Mar 2 18:50:46 UTC 2022 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='2845584122'
GIT_COMMIT_SHA='ee4b93022a5aae3346ff3ebce2d4ed2dbe8ba82c'
GIT_BRANCH='main'



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import Container
from typing import Optional

from vdk.api.plugin.hook_markers import hookspec
from vdk.internal.builtin_plugins.connection.decoration_cursor import DecorationCursor
from vdk.internal.builtin_plugins.connection.execution_cursor import (
    ExecuteOperationResult,
)
from vdk.internal.builtin_plugins.connection.execution_cursor import ExecutionCursor
from vdk.internal.builtin_plugins.connection.recovery_cursor import RecoveryCursor


class ConnectionHookSpec:
    """
    These are hook specifications that enable plugins to hook
        at PEP249Connection and PEP249Cursor events during execution.
    Sequence of evaluation:
        1. db_connection_validate_operation
        2. db_connection_decorate_operation
        * cursor is executed
            ** in case of recovery needed -> db_connection_recover_operation

    Use Cases
        db_connection_validate_operation: if valid return  else raise exception
        db_connection_decorate_operation: ..set_operation( prefix + operation)
        db_connection_recover_operation: ..get_retries()..retry_operation()..
    """

    @hookspec
    def db_connection_validate_operation(
        self, operation: str, parameters: Optional[Container]
    ) -> None:
        """
        Validates the query and parameters.

        For example:
        @hookimpl
        db_connection_validate_operation(operation, parameters):
            if(_max_query_length_limit_exceeded(operation, parameters)):
                raise Exception("max query length limit exceeded")

        :param operation: str
            Database operation - a SQL query or command to be executed.
        :param parameters: Optional[Container]
            Parameters may be provided as sequence or mapping and will be bound to variables in the operation.
            Variables are specified in a database-specific notation. See chosen database documentation for details.
        :return:
        """
        pass

    @hookspec
    def db_connection_decorate_operation(
        self, decoration_cursor: DecorationCursor
    ) -> None:
        """
        Curates the operation and parameters.
        If an exception is raised, the cursor execution will be suspended.

        For example:
        @hookimpl
        db_connection_decorate_operation(decoration_cursor: DecorationCursor):
            managed_operation = decoration_cursor.get_managed_operation()
            managed_operation.set_operation("prefix" +
                                             managed_operation.get_operation())

        :param decoration_cursor: DecorationCursor
            A PEP249Cursor implementation purposed for query and parameters decoration.
            Provides operation details and tooling.
        :return:
        """
        pass

    @hookspec(firstresult=True)
    def db_connection_execute_operation(
        self, execution_cursor: ExecutionCursor
    ) -> Optional[ExecuteOperationResult]:
        """
        The method that executes the actual SQL query using execution cursor.

        When the hook is called the call loop will only invoke up to
        the first hookimpl which returns a result other then None.
        It is then taken as result of the overall hook call.

        You can see the default implementation at DefaultConnectionHookImpl.

        If you want to override the default implementation return a non-None result using tryfirst=True or no decorator.
        But in most cases you would not need to overwrite the default implementation (hence return None).
        If some hookimpl function raises an exception, the behaviour is as outlined in hook_markers module

        Often it may be needed to wrap around it so that you can track the execution of the query.

        For example:

        .. code-block::

            @hookimpl(hookwrapper=True)
            db_connection_execute_operation(execution_cursor: ExecutionCursor) -> Optional[int]:
                # let's track duration of the query
                start = time.time()
                log.info(f"Starting query: {execution_cursor.get_managed_operation().get_operation()}")
                outcome: HookCallResult
                outcome = yield # we yield the execution to other implementations (including default one)
                is_success: bool = outcome.excinfo is None
                log.info(f"Query finished. duration: {time.time() - start}. is_success: {is_success}")
                # no return!

        Another example - we want to change the result

        .. code-block::

            @hookimpl(hookwrapper=True)
            db_connection_execute_operation(execution_cursor: ExecutionCursor) -> Optional[int]:
                outcome: HookCallResult
                outcome = yield #
                outcome.force_result(new_result)  # set new return result

        Another example - let's say we are writing vdk-impala plugin and want to print more debug info
        which is available from the Impala native cursor (provided by impyla library)

        .. code-block::

            @hookimpl(hookwrapper=True)
            db_connection_execute_operation(execution_cursor: ExecutionCursor) -> Optional[int]:
                yield # let the query execute first
                c = cast(impala.interface.Cursor, execution_cursor)
                log.info(f"Query {execution_cursor.get_managed_operation().get_operation()} debug info:"
                         f"summary: {c.get_summary()}, profile: {c.get_profile()}")


        :param execution_cursor: ExecutionCursor
            A PEP249Cursor implementation purposed for actual query execution.
        """
        pass

    @hookspec
    def db_connection_recover_operation(self, recovery_cursor: RecoveryCursor) -> None:
        """
        Recovers the operation initiated. Retries made number is auto-incremented.
        If no exception is raised, cursor is considered successfully executed.

        For example:
        @hookimpl
        db_connection_recover_operation(recovery_cursor: RecoveryCursor):
            while recovery_cursor.get_retries() < MAX_RETRIES:
                try:
                    recovery_cursor.execute("helper query")
                    recovery_cursor.retry_operation()
                    return
                except:
                    time.sleep(3 * recovery_cursor.get_retries()) # backoff
            raise recovery_cursor.get_exception()

        :param recovery_cursor: RecoveryCursor
            A PEP249Cursor implementation purposed for query and parameters recovery.
            Provides operation details and tooling.
        :return:
        """
        pass



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import sys
from typing import cast
from typing import List

import click
import click_log
from click_plugins import with_plugins
from pkg_resources import iter_entry_points
from vdk.api.plugin.core_hook_spec import CoreHookSpecs
from vdk.api.plugin.hook_markers import hookimpl
from vdk.api.plugin.plugin_registry import IPluginRegistry
from vdk.internal.builtin_plugins import builtin_hook_impl
from vdk.internal.builtin_plugins.internal_hookspecs import InternalHookSpecs
from vdk.internal.core.config import Configuration
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.context import CoreContext
from vdk.internal.core.statestore import StateStore
from vdk.internal.plugin.plugin import PluginRegistry

log = logging.getLogger(__name__)


# TODO: perhaps we do not need click-plugins and we can use vdk_initialize hook (and cli.add_command)
@with_plugins(iter_entry_points("vdk.plugin.cli"))
@click.group(
    help="""Command line tool for Data Jobs management.

The cli enables you to conveniently create, develop, run, deploy, list and manage Data Jobs.

\b
# Run to see list of available commands
vdk --help

\b
# Show help of run command (can be done for each command)
vdk run --help

"""
)
@click_log.simple_verbosity_option(logging.getLogger())
@click.pass_context
def cli(ctx: click.Context) -> int:
    """
    Method used to declare root CLI command through decorators.
    """
    return 0


def build_configuration(plugin_registry) -> Configuration:
    """
    We are calling vdk_configure hook and building Configuration of the CLI.
    """
    log.debug("Configure ...")
    conf_builder = ConfigurationBuilder()
    cast(CoreHookSpecs, plugin_registry.hook()).vdk_configure(
        config_builder=conf_builder
    )
    configuration = conf_builder.build()
    return configuration


def setup_cli_commands(plugin_registry: IPluginRegistry, root_command: click.Group):
    """
    Call hooks to customize CLI Commands
    """
    log.debug("Setup commands and options ...")
    cast(CoreHookSpecs, plugin_registry.hook()).vdk_command_line(
        root_command=root_command
    )


def build_core_context_and_initialize(
    configuration: Configuration, plugin_registry: IPluginRegistry
) -> CoreContext:
    core_context = CoreContext(plugin_registry, configuration, StateStore())
    log.debug("Initialize ...")
    cast(CoreHookSpecs, plugin_registry.hook()).vdk_initialize(context=core_context)
    return core_context


class CliEntry:
    @hookimpl(trylast=True)
    def vdk_cli_execute(
        self,
        root_command: click.Command,
        command_line_args: List,
        program_name: str,
        core_context: CoreContext,
    ) -> int:
        # arguments passed are propagated to click.core.main
        exit_code = root_command(
            args=command_line_args,
            prog_name=program_name,
            complete_var=None,
            standalone_mode=True,
            obj=core_context,
        )
        return exit_code

    @hookimpl(trylast=True)
    def vdk_main(
        self,
        plugin_registry: IPluginRegistry,
        root_command: click.Group,
        command_line_args: List,
    ) -> int:
        """
        Main method of the CLI. It call all configuration and initialization hooks and start CLI
        """

        plugin_registry.add_hook_specs(CoreHookSpecs)
        plugin_registry.load_plugin_with_hooks_impl(builtin_hook_impl, "core-plugin")

        program_name = "vdk"  # TODO: infer
        plugin_registry.hook().vdk_start.call_historic(
            kwargs=dict(
                plugin_registry=plugin_registry, command_line_args=command_line_args
            )
        )

        setup_cli_commands(plugin_registry, root_command)
        configuration = build_configuration(plugin_registry)
        core_context = build_core_context_and_initialize(configuration, plugin_registry)

        exit_code = 0
        try:
            log.info(f"Start CLI {program_name} with args {command_line_args}")
            exit_code = cast(InternalHookSpecs, plugin_registry.hook()).vdk_cli_execute(
                root_command=root_command,
                command_line_args=command_line_args,
                core_context=core_context,
                program_name=program_name,
            )
            return exit_code
        except Exception as e:
            handled = cast(CoreHookSpecs, plugin_registry.hook()).vdk_exception(
                exception=e
            )
            # if at least one hook implementation returned handled, means we do not need to log the exception
            if not (True in handled):
                log.exception("Exiting with exception.")
                exit_code = 1
            else:
                exit_code = 0
            return exit_code
        finally:
            cast(CoreHookSpecs, plugin_registry.hook()).vdk_exit(
                context=core_context, exit_code=exit_code
            )


def main() -> None:
    """
    This the starting point for the Python vdk console script.
    """
    # configure basic logging , it's expected that a plugin would override and set it up properly
    click_log.basic_config(logging.getLogger())

    log.debug("Setup plugin registry and call vdk_start hooks ...")
    plugin_registry = PluginRegistry()
    plugin_registry.add_hook_specs(InternalHookSpecs)
    plugin_registry.load_plugins_from_setuptools_entrypoints()
    plugin_registry.load_plugin_with_hooks_impl(CliEntry(), "cli-entry")

    exit_code = cast(InternalHookSpecs, plugin_registry.hook()).vdk_main(
        plugin_registry=plugin_registry,
        root_command=cli,
        command_line_args=sys.argv[1:],
    )
    sys.exit(exit_code)


if __name__ == "__main__":
    main()
